from enum import Enum
from typing import Any, Callable, Awaitable
from collections import defaultdict
from .const import WebsocketType
from .utils import EventItem

import bson  # This is provided by pymongo
import aiohttp
import logging
import asyncio


class Messages(Enum):
    TUNNEL_SERVER_TO_CLIENT_NEW_CONNECTION = 0  # Request from server to client to open a new websocket since there is a new connection to the tunnel.
    DELETE_TUNNEL = 1  # Request from the server to client to delete a tunnel on the client side

##### Serialize and Deserialize Messages bson support #####

MESSAGES_TYPES_KEY = "__Messages__"
def _encode_message_data(data: dict):
    """
    Recursively encoding the data to support Messages objects
    """
    result = {}
    for key, value in data.items():
        if isinstance(value, Messages):
            result[key] = {MESSAGES_TYPES_KEY: value.name}
        elif isinstance(value, dict):
            result[key] = _encode_message_data(value)
        else:
            result[key] = value
    return result

def _decode_message_data(data: dict):
    """
    Recursively decoding the data to support Messages objects
    """
    result = {}
    for key, value in data.items():
        if isinstance(value, dict):
            if MESSAGES_TYPES_KEY in value:
                result[key] = getattr(Messages, value[MESSAGES_TYPES_KEY])
            else:
                result[key] = _decode_message_data(value)
        else:
            result[key] = value
    return result

def message_bson_dumps(data: dict) -> bytes:
    """
    Return a bytes dump of data using bson.encode with support to Message objects
    """
    encoded_data = _encode_message_data(data)
    return bson.encode(encoded_data)

def message_bson_loader(data: bytes) -> dict:
    """
    Return a dictionary of data using bson.loads with support to Message objects
    """
    data = bson.decode(data)
    return _decode_message_data(data)

################################################################


class ChannelError(Exception):
    pass


class ChannelMessage:
    def __init__(self, message_type: Messages, data: dict = None, _identifier: int = None):
        """
        Represent a request over the control channel. Each request has a type, and optionally additional data.
        The identifier is used to match ChannelMessage with a ChannelResponse as request <-> response relationship
        :param message_type: Messages object
        :param data: data to send. Must be serializable to json.
        :param _identifier: This is used internally
        """
        self.message_type = message_type
        self.data = data or {}
        self._identifier = _identifier
        self.channel: Channel = None

    def get_error_response(self, err_message, data: Any = None) -> 'ChannelResponse':
        """
        Generates an ChannelResponse for this message with error message
        """
        if self._identifier is None:
            raise ChannelError('Missing _identifier to generate a response. This is probably a bug')
        return ChannelResponse(message_identifier=self._identifier, data=data, error=err_message)

    def get_valid_response(self, data: Any = None) -> 'ChannelResponse':
        """
        Generates and ChannelResponse for this message the the given data
        """
        if self._identifier is None:
            raise ChannelError('Missing _identifier to generate a response. This is probably a bug')
        return ChannelResponse(message_identifier=self._identifier, data=data)

    def to_dict(self):
        return {
            '_type': 'message',
            '_identifier': self._identifier,
            'message_type': self.message_type,
            'data': self.data
        }

    @staticmethod
    def from_dict(message) -> 'ChannelMessage':
        return ChannelMessage(**message)


class ChannelResponse:
    def __init__(self, message_identifier: int, data: dict = None, error: Any = None):
        """
        Represent a response over the control channel. A response is always associated with an ChannelMessage identifier.
        Do not initialize it directly, use ChannelMessage.get_error_response and ChannelMessage.get_valid_response instead
        """
        self._message_identifier = message_identifier
        self.data = data or {}
        self.error = error

    def is_ok(self):
        return self.error is None

    def to_dict(self):
        response = {
            '_type': 'response',
            '_identifier': self._message_identifier,
            'data': self.data
        }
        if self.error is not None:
            response['error'] = self.error
        return response

    @staticmethod
    def from_dict(response) -> 'ChannelResponse':
        message_identifier = response.pop('_identifier')
        return ChannelResponse(message_identifier=message_identifier, **response)


class Channel:
    def __init__(self, websocket: WebsocketType, channel_id: int, handler: 'Callable[[ChannelMessage], Awaitable[ChannelResponse]]' = None, logger: logging.Logger = None):
        """
        Creates a link between 2 peers on top of a websocket.
        Websockets communication is bi-directional, each message is independent so we can't tell which message is a request and
        which is a response. This creates a protocol on top of a websocket to solve this problem.
        :param websocket: An established websocket between 2 servers. Channel is responsible for closing it
        :param handler: an awaitable callable to handle ChannelMessages. If left as None, we won't handle incoming messages, but we'll still be able to send
        :param channel_id: a unique id for that Channel.
        :param logger: additional logger
        """
        self._websocket = websocket
        self._handler = handler
        self._id = channel_id
        self._logger = logger or logging.getLogger(__name__)
        self._message_id_counter = 0
        # A dictionary of {message_id: Event}. This dict register message_ids which awaits for response.
        # When a new ChannelResponse arrives, if it has a message_id in this dict, we use the event to notify the subscriber.
        self._subscribed_messages = defaultdict(EventItem)
        # This lock is used when registering new message on self._subscribed_messages to prevent duplications of
        # message_ids in case multiple messages received at once.
        self._subscribe_messages_lock = asyncio.Lock()

    @property
    def id(self):
        return self._id

    @property
    def running(self):
        return not self._websocket.closed

    async def send_message(self, message: 'ChannelMessage', timeout=None, raise_if_error=False) -> 'ChannelResponse':
        """
        Send an ChannelMessage to the remote peer and wait for a response.
        """
        if self._websocket.closed:
            raise ChannelError('Channel is closed')
        # Subscribe the message so we'll be notified when a response come back
        async with self._subscribe_messages_lock:
            self._message_id_counter += 1
            message_id = self._message_id_counter
            self._subscribed_messages[message_id].clear()

        # Mark the message_id before sending it
        message_payload = message.to_dict()
        message_payload['_identifier'] = message_id
        try:
            await self._websocket.send_bytes(message_bson_dumps(message_payload))
            # wait for the response to arrive
            wait_coro = self._subscribed_messages[message_id].wait()
            if timeout:
                wait_coro = asyncio.wait_for(wait_coro, timeout=timeout)
            response: ChannelResponse = await wait_coro
        finally:
            del self._subscribed_messages[message_id]
        if raise_if_error:
            if not response.is_ok():
                raise ChannelError(response.error)
        return response

    async def serve(self):
        """
        Start listening for incoming traffic from the websocket.
        """
        async for msg in self._websocket:
            msg: aiohttp.WSMessage
            if msg.type != aiohttp.WSMsgType.BINARY:
                # Websocket related traffic. We ignore it
                self._logger.debug('Channel `%s` Received message of type `%s` with data `%s`. Ignoring', self._id, msg.type, msg.data)
                continue
            message = message_bson_loader(msg.data)
            message_type = message.pop('_type')
            if message_type == 'message':
                # This is a message. We'll handle it and return a response. If no handler, a valid response will return
                message = ChannelMessage.from_dict(message)
                if self._handler:
                    # Inject Channel to the ChannelMessage so that the handler can access it
                    message.channel = self
                    try:
                        response = await self._handler(message)
                    except Exception as err:
                        self._logger.exception('Got exception while handling channel message of type `%s` for channel `%s`:', message.message_type.name, self.id)
                        response = message.get_error_response(err_message=str(err))
                else:
                    response = message.get_valid_response()
                await self._websocket.send_bytes(message_bson_dumps(response.to_dict()))
            elif message_type == 'response':
                message_identifier = message['_identifier']
                if message_identifier in self._subscribed_messages:
                    # We'll pass the message to the subscriber
                    response = ChannelResponse.from_dict(message)
                    self._subscribed_messages[message_identifier].set(response)
                else:
                    self._logger.warning('No subscriber for the message: %s', message)
            else:
                self._logger.warning('Unknown message type: %s', message_type)
        if self._websocket.exception():
            raise self._websocket.exception()
        try:
            close_reason = aiohttp.WSCloseCode(self._websocket.close_code)
            if close_reason is not aiohttp.WSCloseCode.OK:
                self._logger.debug('Channel `%s` closed unexpectedly with close code: %s(%s)', self.id, close_reason.value, close_reason.name)
        except ValueError:
            self._logger.debug('Channel `%s` closed unexpectedly with unknown close code: %s', self.id, self._websocket.close_code)

    async def close(self):
        """
        Closes the websocket, and therefore the channel itself.
        """
        await self._websocket.close()


from typing import Any, Union, Iterable
from aiohttp.web import WebSocketResponse
from aiohttp import ClientWebSocketResponse, ClientSession, BasicAuth, ClientError, TCPConnector

from .exceptions import NETunnelInvalidProxy

import logging
import socket
import asyncio
import contextlib
import functools
import sys
import http

LOGGING_FORMATTER = '%(asctime)s %(name)s - %(levelname)s - %(message)s'


def asyncio_current_task(loop=None):
    if sys.version_info >= (3, 9, 0):
        return asyncio.current_task(loop)
    return asyncio.Task.current_task(loop)


def asyncio_all_tasks(loop=None):
    if sys.version_info >= (3, 9, 0):
        return asyncio.all_tasks(loop)
    return asyncio.Task.all_tasks(loop)


async def get_session(headers=None, ssl=None) -> ClientSession:
    """
    Return an aiohttp.ClientSession object.
    ClientSession must be initialized inside an async context.
    :param headers: Optional headers to append for every request that this session will make
    :param ssl: SSLContext object. False to skip validation, None for default SSL check.
    """
    connector = TCPConnector(ssl=ssl)
    headers = headers or {}
    return ClientSession(connector=connector, ws_response_class=EventClientWebSocketResponse, headers=headers)


def get_logger(name, level=logging.INFO):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(LOGGING_FORMATTER)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


async def run_blocking_func_in_executor(func, *args, **kwargs):
    """
    Run a blocking function in a different executor so that it won't stop
    the event loop
    """
    return await asyncio.get_event_loop().run_in_executor(None, functools.partial(func, *args, **kwargs))


def get_unused_port(min_port: int, max_port: int, exclude_ports: Iterable = None, address='127.0.0.1'):
    """
    Return an unused port from a range of integers
    :param min_port: minimum port that can be allocated
    :param max_port: maximum port that can be allocated
    :param exclude_ports: list of ports to exclude
    :param address: address of the interface to try binding with
    """
    exclude_ports = set(exclude_ports) or set()
    for port in range(min_port, max_port):
        if port in exclude_ports:
            continue
        with contextlib.suppress(OSError):
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind((address, port))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                return port
    raise RuntimeError(f"Failed to found an available port between {min_port} and {max_port}")


@contextlib.contextmanager
def object_in_list(obj: Any, list_of_obj: list):
    """
    append obj to list_of_obj and remove it once the with statement finishes.
    Used like this:

    with object_in_list(item, list_of_items):
        # item is in list_of_items
    # item is not in list_of_items
    """
    list_of_obj.append(obj)
    try:
        yield obj
    finally:
        with contextlib.suppress(ValueError):
            list_of_obj.remove(obj)


def task_in_list_until_done(task: Union[asyncio.Task, asyncio.Future], list_of_tasks: list):
    """
    Add a task to a list and a callback which removes it from the list once it's done
    """
    def remove_task(*args):
        with contextlib.suppress(ValueError):
            list_of_tasks.remove(task)
    list_of_tasks.append(task)
    task.add_done_callback(remove_task)


def update_dict_recursively(dict_to_update: dict, dict_new_values: dict):
    """
    Update the target dictionary recursively
    """
    for key, value in dict_new_values.items():
        if isinstance(value, dict):
            # If the target corresponding key is not a dict as well, we'll override it
            # because it's irrelevant and then we can use update_dict_recursively again
            if key not in dict_to_update or not isinstance(dict_to_update[key], dict):
                dict_to_update[key] = {}
            update_dict_recursively(dict_to_update[key], value)
        else:
            dict_to_update[key] = value


class EventItem(asyncio.Event):
    def __init__(self, loop=None):
        """
        Works just like asyncio.Event with the following enhancements:
        - Stores an item when setting the event and retrieve it when it's available with self.wait
        """
        super().__init__(loop=loop)
        self._obj: Any = None

    def set(self, obj: Any = None):
        self._obj = obj
        return super().set()

    def clear(self):
        self._obj = None
        return super().clear()

    async def wait(self) -> Any:
        await super().wait()
        return self._obj


class EventQueue(asyncio.Queue):
    """
    This queue works just like a normal async queue, a new method self.join_no_empty
    was added. It works just like self.join but in opposite, it returns whenever
    a new item was put in the queue without extracting it
    """
    def __init__(self, maxsize=0):
        super().__init__(maxsize=maxsize)
        self._queue_no_empty = asyncio.Event()

    async def put(self, item):
        await super().put(item)
        self._queue_no_empty.set()

    async def get(self):
        try:
            return await super().get()
        finally:
            if self.qsize() == 0:
                self._queue_no_empty.clear()

    def get_nowait(self):
        try:
            return super().get_nowait()
        finally:
            if self.qsize() == 0:
                self._queue_no_empty.clear()

    async def join_no_empty(self):
        """
        Blocks until a new task is in the queue
        """
        await self._queue_no_empty.wait()


class EventWebSocketResponse(WebSocketResponse):
    """
    Works the same as WebSocketResponse with new methods:
    - wait_closed - block until websocket is close (does not trigger close)
    - close_nowait - trigger close websocket without blocking
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._shutdown = asyncio.Event()
        self._closing_task: asyncio.Future = None

    async def wait_closed(self):
        """
        Block until the websocket is closed
        """
        await self._shutdown.wait()
        if self._closing_task is not None:
            with contextlib.suppress(asyncio.CancelledError):
                await self._closing_task

    def close_nowait(self):
        """
        Closing without waiting.
        Use wait_closed to await for the close task
        """
        self._closing_task = asyncio.ensure_future(self.close())

    async def close(self, *args, **kwargs):
        return_value = await super().close(*args, **kwargs)
        self._shutdown.set()
        return return_value


class EventClientWebSocketResponse(ClientWebSocketResponse):
    """
    Works the same as ClientWebSocketResponse with new methods:
    - wait_closed - block until websocket is close (does not trigger close)
    - close_nowait - trigger close websocket without blocking
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._shutdown = asyncio.Event()
        self._closing_task: asyncio.Future = None

    async def wait_closed(self):
        """
        Block until the websocket is closed
        """
        await self._shutdown.wait()
        if self._closing_task is not None:
            with contextlib.suppress(asyncio.CancelledError):
                await self._closing_task

    def close_nowait(self):
        """
        Closing without waiting.
        Use wait_closed to await for the close task
        """
        self._closing_task = asyncio.ensure_future(self.close())

    async def close(self, *args, **kwargs):
        return_value = await super().close(*args, **kwargs)
        self._shutdown.set()
        return return_value


async def verify_proxy(proxy_url, test_url, username=None, password=None):
    """
    Verify that a proxy server is working
    :param proxy_url: url to the proxy
    :param test_url: url that accepts GET method and will be used to test the proxy
    :param username: username to use to authenticate the proxy
    :param password: password to use to authenticate the proxy
    """
    proxy_auth = None
    if username and password:
        proxy_auth = BasicAuth(username, password)
    elif (username and password is None) or (password and username is None):
        raise ValueError('HTTP proxy authentication should include both username and password')
    async with ClientSession() as client:
        try:
            async with client.get(test_url, proxy=proxy_url, proxy_auth=proxy_auth) as resp:
                if resp.status != http.HTTPStatus.OK.value:
                    raise NETunnelInvalidProxy(f"Invalid response from proxy: {resp.status}")
        except (ClientError, ConnectionRefusedError) as err:
            raise NETunnelInvalidProxy(str(err))



import http


class NETunnelError(Exception):
    pass


class NETunnelNotConnectedError(NETunnelError):
    pass


class NETunnelNotAuthenticatedError(NETunnelError):
    pass


class NETunnelResponseError(NETunnelError):
    pass


class NETunnelDestinationNotAllowed(NETunnelError):
    pass


class NETunnelInvalidProxy(NETunnelError):
    pass


class NETunnelServerError(NETunnelError):
    """
    Used by the server to raise exception and convert them to json responses with a specific http status code bound to them
    """
    status_code = http.HTTPStatus.INTERNAL_SERVER_ERROR


class NETunnelServerNotFound(NETunnelServerError):
    status_code = http.HTTPStatus.NOT_FOUND


class NETunnelServerBadRequest(NETunnelServerError):
    status_code = http.HTTPStatus.BAD_REQUEST


class NETunnelAuthError(NETunnelServerError):
    status_code = http.HTTPStatus.FORBIDDEN



"""
Security utilities for netunnel
"""
from cryptography.fernet import Fernet

import click


@click.group(help="Toolkit for security utilities on netunnel")
def main():
    pass


@main.command(name='generate-key')
def _generate_key():
    """Generates a key that can be used on NETunnelServer
    """
    print(Encryptor.generate_key().decode())


@main.command(name='encrypt')
@click.argument('secret-key')
@click.argument('data')
def _encrypt(secret_key, data):
    """Encrypt data using the secret_key"""
    print(Encryptor(secret_key).encrypt_string(data))


@main.command(name='decrypt')
@click.argument('secret-key')
@click.argument('data')
def _decrypt(secret_key, data):
    """Decrypt data using the secret_key"""
    print(Encryptor(secret_key).decrypt_string(data))


class Encryptor:
    def __init__(self, key):
        self._fernet = Fernet(key)

    @staticmethod
    def generate_key():
        return Fernet.generate_key()

    def encrypt(self, data: bytes):
        return self._fernet.encrypt(data)

    def decrypt(self, data: bytes):
        return self._fernet.decrypt(data)

    def encrypt_string(self, data: str):
        return self.encrypt(data.encode()).decode()

    def decrypt_string(self, data: str):
        return self.decrypt(data.encode()).decode()


if __name__ == '__main__':
    main()


from marshmallow import Schema, fields, validate, pre_load, ValidationError

import ipaddress


class NETunnelSchema(Schema):
	"""
	Base schema for netunnel objects

	For backwards compatibility to marshmallow 2.X, we're using self.dump3 and self.load3 instead
	of self.dump & self.load to make them as similar as possible to marshmallow 3.X.
	We're not overriding these methods directly to avoid breaking marshmallow as they're being used internally.
	"""
	def dump3(self, obj, *args, **kwargs):
		dump_result = super().dump(obj, *args, **kwargs)
		# We cannot use isinstance because MarshalResult won't exists on marshmallow 3.X
		if type(dump_result).__name__ == 'MarshalResult':
			if dump_result.errors:
				raise ValidationError(message=dump_result.errors)
			return dump_result.data
		return dump_result

	def load3(self, obj, *args, **kwargs):
		load_result = super().load(obj, *args, **kwargs)
		# We cannot use isinstance because UnmarshalResult won't exists on marshmallow 3.X
		if type(load_result).__name__ == 'UnmarshalResult':
			if load_result.errors:
				raise ValidationError(message=load_result.errors)
			return load_result.data
		return load_result


class StaticTunnelSchema(NETunnelSchema):
	id = fields.Integer()
	tunnel_remote_address = fields.String(default='127.0.0.1', missing='127.0.0.1')
	tunnel_remote_port = fields.Integer(required=True, validate=validate.Range(min=1, max=65535))
	tunnel_local_address = fields.String(default='127.0.0.1', missing='127.0.0.1')
	tunnel_local_port = fields.Integer(required=True, validate=validate.Range(min=1, max=65535))

	@pre_load
	def validate_scheme(self, data, **kwargs):
		for address in ['tunnel_remote_address', 'tunnel_local_address']:
			try:
				ipaddress.IPv4Address(data.get(address, '127.0.0.1'))
			except ipaddress.AddressValueError as err:
				raise ValidationError(str(err))
		return data


class PeerSchema(NETunnelSchema):
	id = fields.Integer()
	name = fields.String(required=True)
	target_netunnel_url = fields.URL(required=True)
	auth_data = fields.Dict(required=True, keys=fields.String(), values=fields.String())
	static_tunnels = fields.List(fields.Nested(StaticTunnelSchema))



from ..client import NETunnelClient
from ..common.tunnel import Tunnel
from ..common.utils import get_logger
from ..common.exceptions import NETunnelAuthError

import time
import aiohttp
import logging
import asyncio
import contextlib

CONNECTION_RETRY_INTERVAL = 10
HEALTH_CHECK_INTERVAL = 30


class StaticTunnel:
    def __init__(self, id, target_netunnel_url, tunnel_local_address, tunnel_local_port, tunnel_remote_address,
                 tunnel_remote_port, auth_client, proxy_url=None, proxy_username=None, proxy_password=None, ssl=None,
                 logger=None, connection_retry_interval=CONNECTION_RETRY_INTERVAL):
        """
        Interface for managing a static tunnel
        :param id: unique id for that static tunnel
        :param target_netunnel_url: url of the remote peer's netunnel server
        :param tunnel_local_address: local address to use for the static tunnel
        :param tunnel_local_port: local port to use for the static tunnel
        :param tunnel_remote_address: remote address to use for the static
        :param tunnel_remote_port: remote port to use for the static tunnel
        :param auth_client: Instance of subclass of netunnel.common.auth.NETunnelClientAuth that will be used to authenticate the remote
        :param proxy_url: url for a proxy to use when making requests to the remote netunnel server
        :param proxy_username: Optional username to use when authenticating with the proxy_url. `proxy_password` must be given as well
        :param proxy_password: Optional password to use when authenticating with the proxy_url. `proxy_username` must be given as well
        :param ssl: SSLContext object. False to skip validation, None for default SSL check.
        :param logger: logger to use
        :param connection_retry_interval: interval for retrying to connect the remote peer after failure
        """
        self._id = id
        self._client: NETunnelClient = NETunnelClient(server_url=target_netunnel_url, proxy_url=proxy_url,
                                                        proxy_username=proxy_username, proxy_password=proxy_password,
                                                        auth_client=auth_client, ssl=ssl, logger=logger)
        self._tunnel_remote_address = tunnel_remote_address
        self._tunnel_remote_port = tunnel_remote_port
        self._tunnel_local_address = tunnel_local_address
        self._tunnel_local_port = tunnel_local_port
        self._tunnel: Tunnel = None
        # _running is True when the static tunnel needs to be online
        self._running = False
        # This event is True when the static tunnel is actually online
        self._online_event = asyncio.Event()
        self._logger: logging.Logger = logger or get_logger(f'static_tunnel `{self._id}`')
        self._tunnel_running_task: asyncio.Future = None
        self._tunnel_watchdog_task: asyncio.Future = None
        self._connection_retry_interval = connection_retry_interval

    @property
    def id(self):
        return self._id

    @property
    def tunnel_local_address(self):
        return self._tunnel_local_address

    @property
    def tunnel_local_port(self):
        return self._tunnel_local_port

    @property
    def tunnel_remote_address(self):
        return self._tunnel_remote_address

    @property
    def tunnel_remote_port(self):
        return self._tunnel_remote_port

    @property
    def running(self):
        return self._running

    def get_tunnel_display_name(self):
        return f'local on {self._tunnel_local_address}:{self._tunnel_local_port} -> {self._client.server_url} on {self._tunnel_remote_address}:{self._tunnel_remote_port}'

    async def start(self):
        """
        Start connecting the tunnel to the peer. If peer is not available
        or the connection is lost, keep retrying
        """
        if self._running:
            raise RuntimeError('Static tunnel already started')
        self._running = True
        self._tunnel_running_task = asyncio.ensure_future(self._tunnel_keep_alive())
        self._tunnel_watchdog_task = asyncio.ensure_future(self._tunnel_watchdog())

    async def wait_online(self):
        """
        Blocks until the the static tunnel is up and running
        """
        await self._online_event.wait()

    async def set_new_proxy(self, proxy_url, proxy_username, proxy_password):
        """
        Set a new http proxy to the client and close the existing tunnel so it will recreate it with the new proxy
        """
        if not self._running:
            self._client.set_client_proxy(proxy_url, proxy_username, proxy_password)
            return
        try:
            await self.stop()
            self._client.set_client_proxy(proxy_url, proxy_username, proxy_password)
        finally:
            await self.start()

    async def _tunnel_watchdog(self):
        """
        Perform health checks to the tunnel every certain interval.
        if the tunnel marked as running but the health check fails, we stop it so that
        the _tunnel_keep_alive task can wake up and restart it
        """
        while self._running:
            if self._tunnel and self._tunnel.running and not await self._tunnel.health_check():
                # We use force because the tunnel is probably malfunctioning and might hang with graceful shutdown
                self._logger.info('Static Tunnel `%s` is not working. Restarting tunnel', self.get_tunnel_display_name())
                await self._tunnel.stop(force=True)
            await asyncio.sleep(HEALTH_CHECK_INTERVAL)

    async def _tunnel_keep_alive(self):
        """
        try to connect to the peer and join the tunnel. If peer disconnected or unavailable,
        keep trying until closed.
        """
        while self._running:
            try:
                async with self._client:
                    start_time = time.time()
                    self._tunnel: Tunnel = await asyncio.wait_for(self._client.open_tunnel_to_server(remote_address=self._tunnel_remote_address,
                                                                                                     remote_port=self._tunnel_remote_port,
                                                                                                     local_address=self._tunnel_local_address,
                                                                                                     local_port=self._tunnel_local_port,
                                                                                                     wait_ready=True),
                                                                  self._connection_retry_interval)
                    self._logger.info('Static Tunnel `%s` is online', self.get_tunnel_display_name())
                    self._online_event.set()
                    await self._tunnel.join()
                self._online_event.clear()
                # At this point, the tunnel might have been closed unexpectedly so we break only if we're signaled to stop
                if self._running is False:
                    self._logger.info('Static Tunnel `%s` is closed', self.get_tunnel_display_name())
                    break
                self._logger.info('Static Tunnel `%s` is offline after %s seconds. Waiting %s seconds before reconnecting...',
                                  self.get_tunnel_display_name(), int(time.time() - start_time), self._connection_retry_interval)
                await asyncio.sleep(self._connection_retry_interval)
            except Exception as err:
                self._online_event.clear()
                if isinstance(err, aiohttp.ClientConnectionError):
                    self._logger.warning('Failed to connect to `%s`', self._client.server_url)
                elif isinstance(err, NETunnelAuthError):
                    self._logger.warning('Failed to authenticate with `%s`', self._client.server_url)
                else:
                    self._logger.exception('Failed to establish a tunnel with the peer `%s`:', self._client.server_url)
                self._logger.info('Waiting %s seconds before retrying...', self._connection_retry_interval)
                await asyncio.sleep(self._connection_retry_interval)


    async def stop(self, force=True):
        """
        Close the connection to the peer
        :param force: Whether to close the tunnel forcefully. Defaults to True
        """
        self._running = False
        if self._tunnel is not None and self._tunnel.running:
            await self._tunnel.stop(force=force)
            await self._tunnel_running_task
        if not self._tunnel_watchdog_task.done():
            # We cancel the watchdog task to avoid waiting for the sleep interval
            self._tunnel_watchdog_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._tunnel_watchdog_task
        await self._client.close()